#ifndef TYPES_H
#define TYPES_H 1

#include <stdbool.h>
#include <stdint.h>

typedef struct _ImlibLoader ImlibLoader;

typedef struct _ImlibImage ImlibImage;

typedef int     ImlibOp;

typedef struct _ImlibColorModifier ImlibColorModifier;

typedef struct _ImlibUpdate ImlibUpdate;

#endif                          /* TYPES_H */
