/*
 * Obsolete, useless code
 */
#include "config.h"
#include <Imlib2.h>

EAPI void
imlib_image_set_irrelevant_border(char irrelevant)
{
}

EAPI void
imlib_image_set_irrelevant_alpha(char irrelevant)
{
}

EAPI void
imlib_context_set_TTF_encoding(Imlib_TTF_Encoding encoding)
{
}

EAPI            Imlib_TTF_Encoding
imlib_context_get_TTF_encoding(void)
{
    return 0;
}
